// demo.cpp : Defines the class behaviors for the application.

#include "stdafx.h"
#include "SVImageViewer.h"

#include "MainFrm.h"
#include "ChildFrm.h"
#include "SVImageViewerDoc.h"
#include "SVImageViewerView.h"
#include "xShadeButton.h"

#include "ximage.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

DocType doctypes[CMAX_IMAGE_FORMATS] =
{
#if CXIMAGE_SUPPORT_BMP
	{ CXIMAGE_FORMAT_BMP, TRUE, TRUE, "BMP files", "*.bmp" },
#endif
#if CXIMAGE_SUPPORT_GIF
	{ CXIMAGE_FORMAT_GIF, TRUE, FALSE,"GIF files", "*.gif" },
#endif
#if CXIMAGE_SUPPORT_JPG
	{ CXIMAGE_FORMAT_JPG, TRUE, TRUE, "JPG files", "*.jpg;*.jpeg" },
#endif
#if CXIMAGE_SUPPORT_PNG
	{ CXIMAGE_FORMAT_PNG, TRUE, TRUE, "PNG files", "*.png" },
#endif
#if CXIMAGE_SUPPORT_MNG
	{ CXIMAGE_FORMAT_MNG, TRUE, TRUE, "MNG files", "*.mng;*.jng;*.png" },
#endif
#if CXIMAGE_SUPPORT_ICO
	{ CXIMAGE_FORMAT_ICO, TRUE, TRUE, "ICO files", "*.ico" },
#endif
#if CXIMAGE_SUPPORT_TIF
	{ CXIMAGE_FORMAT_TIF, TRUE, TRUE, "TIF files", "*.tif;*.tiff" },
#endif
#if CXIMAGE_SUPPORT_TGA
	{ CXIMAGE_FORMAT_TGA, TRUE, TRUE, "TGA files", "*.tga" },
#endif
#if CXIMAGE_SUPPORT_PCX
	{ CXIMAGE_FORMAT_PCX, TRUE, FALSE, "PCX files", "*.pcx" },
#endif
#if CXIMAGE_SUPPORT_SVI
	{ CXIMAGE_FORMAT_SVI, TRUE, TRUE, "SVI ( SuperVozelj Raw 256 Gray Level Bitmap) files", "*.svi" }
#endif

};

//////////////////////////////////////////////////////////////////////////////
// CSVImageViewerApp
BEGIN_MESSAGE_MAP(CSVImageViewerApp, CWinApp)
	//{{AFX_MSG_MAP(CSVImageViewerApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_WINDOW_CLOSEALL, OnWindowCloseall)
	ON_COMMAND(ID_HELP_USESVIMAGEVIEWER, OnHelpUsesvimageviewer)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CSVImageViewerApp::OnFileOpen)
	// Standard print setup command
	ON_COMMAND(ID_FILE_PRINT_SETUP, CWinApp::OnFilePrintSetup)
END_MESSAGE_MAP()

//////////////////////////////////////////////////////////////////////////////
// CSVImageViewerApp construction
CSVImageViewerApp::CSVImageViewerApp()
{
	m_nDocCount=1; //counter for documents
}
//////////////////////////////////////////////////////////////////////////////
// The one and only CSVImageViewerApp object
CSVImageViewerApp theApp;

//////////////////////////////////////////////////////////////////////////////
// CSVImageViewerApp initialization
BOOL CSVImageViewerApp::InitInstance()
{
	// Standard initialization
#ifdef _AFXDLL
	Enable3dControls();			// Call this when using MFC in a shared DLL
#else
	Enable3dControlsStatic();	// Call this when linking to MFC statically
#endif

	LoadStdProfileSettings(9);  // Load standard INI file options (including MRU)

	// Register the application's document templates.  Document templates
	//  serve as the connection between documents, frame windows and views.
	CMultiDocTemplate* pDocTemplate;
	pDocTemplate = new CMultiDocTemplate(
		IDR_SVIMAGEVIEWERTYPE,
		RUNTIME_CLASS(CSVImageViewerDoc),
		RUNTIME_CLASS(CChildFrame), // custom MDI child frame
		RUNTIME_CLASS(CSVImageViewerView));
	demoTemplate = pDocTemplate;
	AddDocTemplate(pDocTemplate);

	// create main MDI Frame window
	CMainFrame* pMainFrame = new CMainFrame;
	if (!pMainFrame->LoadFrame(IDR_MAINFRAME)) return FALSE;
	m_pMainWnd = pMainFrame;

	// This code replaces the MFC created menus with 
	// the Ownerdrawn versions 
	pDocTemplate->m_hMenuShared=pMainFrame->NewMenu();
	pMainFrame->m_hMenuDefault=pMainFrame->NewDefaultMenu();
	// This simulates a window being opened if you don't have
	// a default window displayed at startup
	pMainFrame->OnUpdateFrameMenu(pMainFrame->m_hMenuDefault);

	// Parse command line for standard shell commands, DDE, file open
	CCommandLineInfo cmdInfo;
    // Alter behaviour to not open window immediately
    cmdInfo.m_nShellCommand = CCommandLineInfo::FileNothing;
	ParseCommandLine(cmdInfo);

	// Dispatch commands specified on the command line
	if (!ProcessShellCommand(cmdInfo)) return FALSE;

	// The main window has been initialized, so show and update it.
	pMainFrame->ShowWindow(m_nCmdShow);
	pMainFrame->UpdateWindow();

	// Enable drag/drop open
	m_pMainWnd->DragAcceptFiles();

	// Enable open from command line
	if (m_lpCmdLine != "") OpenDocumentFile(m_lpCmdLine);

	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
// CHelpAboutDlg dialog used for App About
class CHelpAboutDlg : public CDialog
{
public:
	CHelpAboutDlg(int IDD=IDD_ABOUTBOX);
// Dialog Data
	//{{AFX_DATA(CHelpAboutDlg)
	CxShadeButton	m_ok;
	//}}AFX_DATA
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CHelpAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL
// Implementation
protected:
	//{{AFX_MSG(CHelpAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};
//////////////////////////////////////////////////////////////////////////////
CHelpAboutDlg::CHelpAboutDlg(int IDD) : CDialog(IDD)
{
}
//////////////////////////////////////////////////////////////////////////////
void CHelpAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(DlgCapture)
	DDX_Control(pDX, IDOK, m_ok);
	//}}AFX_DATA_MAP
}
//////////////////////////////////////////////////////////////////////////////
BEGIN_MESSAGE_MAP(CHelpAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CHelpAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()
//////////////////////////////////////////////////////////////////////////////
// App command to run the dialog
void CSVImageViewerApp::OnAppAbout()
{
	CHelpAboutDlg aboutDlg;
	aboutDlg.DoModal();
}


//////////////////////////////////////////////////////////////////////////////
void CSVImageViewerApp::OnFileOpen() 
{
	// prompt the user (with all document templates)
	CString newName;
	
	int nDocType = CXIMAGE_FORMAT_SVI;

	if (!PromptForFileName(newName, AFX_IDS_OPENFILE, OFN_HIDEREADONLY | OFN_FILEMUSTEXIST, TRUE, &nDocType))
		return; // open cancelled
	
	OpenDocumentFile(newName);
}
//////////////////////////////////////////////////////////////////////////////
// prompt for file name - used for open and save as
// static function called from app
BOOL CSVImageViewerApp::PromptForFileName(CString& fileName, UINT nIDSTitle, 
	DWORD dwFlags, BOOL bOpenFileDialog, int* pType)
{
	CFileDialog dlgFile(bOpenFileDialog);
	CString title;

	if (bOpenFileDialog) 
		title="Open image file"; 
	else 
		title="Save image file";

	dlgFile.m_ofn.Flags |= dwFlags;

	int nDocType = (pType != NULL) ? *pType : CXIMAGE_FORMAT_JPG;

	int nIndex = GetIndexFromType(nDocType, bOpenFileDialog);

	if (nIndex == -1) 
		nIndex = 0;

	dlgFile.m_ofn.nFilterIndex = nIndex +1;
	// strDefExt is necessary to hold onto the memory from GetExtFromType
	
	CString strDefExt = GetExtFromType(nDocType).Mid(2,3);
	dlgFile.m_ofn.lpstrDefExt = strDefExt;
		
	CString strFilter = GetFileTypes(bOpenFileDialog);
	
	dlgFile.m_ofn.lpstrFilter = strFilter;
	dlgFile.m_ofn.lpstrTitle = title;
	dlgFile.m_ofn.lpstrFile = fileName.GetBuffer(_MAX_PATH);

	BOOL bRet = (dlgFile.DoModal() == IDOK) ? TRUE : FALSE;
	fileName.ReleaseBuffer();
	if (bRet)
	{
		if (pType != NULL)
		{
			int nIndex = (int)dlgFile.m_ofn.nFilterIndex - 1;
			ASSERT(nIndex >= 0);
			*pType = GetTypeFromIndex(nIndex, bOpenFileDialog);
		}
	}
	return bRet;
}
//////////////////////////////////////////////////////////////////////////////
int CSVImageViewerApp::GetIndexFromType(int nDocType, BOOL bOpenFileDialog)
{
	int nCnt = 0;
	for (int i=0;i<CMAX_IMAGE_FORMATS;i++){
		if (bOpenFileDialog ? doctypes[i].bRead : doctypes[i].bWrite){
			if (doctypes[i].nID == nDocType) return nCnt;
			nCnt++;
		}
	}
	return -1;
}
//////////////////////////////////////////////////////////////////////////////
int CSVImageViewerApp::GetTypeFromIndex(int nIndex, BOOL bOpenFileDialog)
{
	int nCnt = 0;
	for (int i=0;i<CMAX_IMAGE_FORMATS;i++){
		if (bOpenFileDialog ? doctypes[i].bRead : doctypes[i].bWrite){
			if (nCnt == nIndex)
//              return i; // PJO - Buglet ?
                return doctypes[i].nID;
			nCnt++;
		}
	}
	ASSERT(FALSE);
	return -1;
}
//////////////////////////////////////////////////////////////////////////////
CString CSVImageViewerApp::GetExtFromType(int nDocType)
{
	for (int i=0;i<CMAX_IMAGE_FORMATS;i++){
		if (doctypes[i].nID == nDocType)
			return doctypes[i].ext;
	}
	return CString("");
}
//////////////////////////////////////////////////////////////////////////////
BOOL CSVImageViewerApp::GetWritableType(int nDocType){
	for (int i=0;i<CMAX_IMAGE_FORMATS;i++){
		if (doctypes[i].nID == nDocType)
			return doctypes[i].bWrite;
	}
	return FALSE;
}
//////////////////////////////////////////////////////////////////////////////
CString CSVImageViewerApp::GetFileTypes(BOOL bOpenFileDialog)
{
	CString str;
	for (int i=0;i<CMAX_IMAGE_FORMATS;i++)
	{
		if (bOpenFileDialog && doctypes[i].bRead)
		{
			str += doctypes[i].description;
			str += (TCHAR)NULL;
			str += doctypes[i].ext;
			str += (TCHAR)NULL;
		} 
		else if (!bOpenFileDialog && doctypes[i].bWrite) 
		{
			str += doctypes[i].description;
			str += (TCHAR)NULL;
			str += doctypes[i].ext;
			str += (TCHAR)NULL;
		}
	}
	
	if (bOpenFileDialog)
	{
		str += "All Files";
		str += (TCHAR)NULL;
		str += "*.*";
		str += (TCHAR)NULL;
	} 

	return str;
}

//////////////////////////////////////////////////////////////////////////////
// CSVImageViewerApp commands
//////////////////////////////////////////////////////////////////////////////
void CSVImageViewerApp::OnWindowCloseall() 
{
	POSITION pos = AfxGetApp()->GetFirstDocTemplatePosition(); 
    while(pos) { 
        CDocTemplate* pTemplate = AfxGetApp()->GetNextDocTemplate(pos); 
        POSITION pos2 = pTemplate->GetFirstDocPosition(); 
        while(pos2) { 
            CDocument* pDocument; 
			if ((pDocument=pTemplate->GetNextDoc(pos2)) != NULL) 
				pDocument->OnCloseDocument();  
		}
	}
}
//////////////////////////////////////////////////////////////////////////////

class CHelpUseDlg : public CDialog
{
// Construction
public:
	CHelpUseDlg(CWnd* pParent = NULL);   // standard constructor

// Dialog Data
	//{{AFX_DATA(CHelpUseDlg)
	enum { IDD = IDD_HELP_USE };
	CxShadeButton	m_ok;
	CString	m_helpUse;
		// NOTE: the ClassWizard will add data members here
	//}}AFX_DATA


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CHelpUseDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:

	// Generated message map functions
	//{{AFX_MSG(CHelpUseDlg)
		// NOTE: the ClassWizard will add member functions here
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CHelpUseDlg::CHelpUseDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CHelpUseDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CHelpUseDlg)
		// NOTE: the ClassWizard will add member initialization here
		m_helpUse = _T("The SuperVozelj file format is a RAW (without header) 256 Gray level bitmap used by SuperVozelj an amateur packet radio controller node to take picture with the onboard CCD camera.Download the image file with your preferred packet radio program and use drag and drop or use menu to load, specify the resolution (default 512x256), modifiy and save it in a different file format. The program automatically try to remove heading and trailing garbage");
	//}}AFX_DATA_INIT
}


void CHelpUseDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CHelpUseDlg)
		// NOTE: the ClassWizard will add DDX and DDV calls here
		DDX_Control(pDX, IDOK, m_ok);
		DDX_Text(pDX, IDC_EDITUSE, m_helpUse);
		DDV_MaxChars(pDX, m_helpUse, 1024);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CHelpUseDlg, CDialog)
	//{{AFX_MSG_MAP(CHelpUseDlg)
		// NOTE: the ClassWizard will add message map macros here
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


void CSVImageViewerApp::OnHelpUsesvimageviewer() 
{
	// TODO: Add your command handler code here
	
	CHelpUseDlg aboutDlg;
	aboutDlg.DoModal();
}
